/* $Id: rss.c 85 2023-05-13 08:55:12Z nishi $ */
#include <swinehub/rss.h>

#include <swinehub/config.h>
#include <swinehub/util.h>
#include <swinehub.h>

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <unistd.h>
#include <fcntl.h>
#include <sys/wait.h>
#include <sys/stat.h>
#include <signal.h>

void swinehub_rss(struct swinehub_config* config){
	char* _cutfrom = strdup(getenv("PATH_REPLACED") + strlen(getenv("FIRST_PATH")) + 1);
	if(strlen(_cutfrom) > 1){
		char* cutfrom = _cutfrom;
		int index = -1;
		cutfrom += 1;
		int i;
		for(i = 0; cutfrom[i]; i++){
			if(cutfrom[i] == '/'){
				index = i;
				cutfrom[i] = 0;
				break;
			}
		}
		if(index == -1){
			printf("Status: 400 Bad Request\r\n");
			printf("Content-Type: text/plain\r\n");
			printf("\r\n");
			printf("Missing required parameters");
		}else{
			char* username = cutfrom;
			char* reponame = cutfrom + i + 1;
			char* checkpath = sstrcat5(config->path, "/", username, "-", reponame);
			struct stat s;
			if(stat(checkpath, &s) != 0){
				free(checkpath);
				free(_cutfrom);
				printf("Status: 404 Not Found\r\n");
				printf("Content-Type: text/plain\r\n");
				printf("\r\n");
				printf("Repository not found\r\n");
				return;
			}
			char* svn_path = sstrcat5(config->cachepath, "/", username, "-", reponame);
			if(stat(svn_path, &s) != 0){
				int pid = fork();
				if(pid == 0){
					int nullfd = open("/dev/null", O_RDWR);
					dup2(nullfd, fileno(stdout));
					dup2(nullfd, fileno(stderr));
					chdir(config->cachepath);
					execl(config->svn, config->svn, "co", sstrcat6("file://", config->path, "/", username, "-", reponame), NULL);
					_exit(-1);
				}else{
					waitpid(pid, 0, 0);
				}
			}
			int pid = fork();
			if(pid == 0){
				int nullfd = open("/dev/null", O_RDWR);
				dup2(nullfd, fileno(stdout));
				dup2(nullfd, fileno(stderr));
				chdir(svn_path);
				execl(config->svn, config->svn, "up", NULL);
				_exit(-1);
			}else{
				waitpid(pid, 0, 0);
			}
			char* revision_file = sstrcat6(config->path, "/", username, "-", reponame, "/db/current");
			FILE* f = fopen(revision_file, "r");
			if(f != NULL){
				printf("Status: 200 OK\r\n");
				printf("Content-Type: application/rss+xml\r\n");
				printf("\r\n");
				fflush(stdout);
				printf("<?xml version=\"1.0\" encoding=\"utf-8\"?>\n");
				printf("<rdf:RDF xmlns:dc=\"http://purl.org/dc/elements/1.1/\" xmlns:rdf=\"http://www.w3.org/1999/02/22-rdf-syntax-ns#\" xmlns=\"http://purl.org/rss/1.0/\">\n");
				printf("	<channel>\n");
				printf("		<title>%s: ~%s/%s</title>\n", config->title, username, reponame);
				printf("		<link>%s%s/~%s/%s</link>\n", config->url_prefix, getenv("SCRIPT_NAME"), username, reponame);
				printf("		<image rdf:resource=\"%s%s\" />\n", config->url_prefix, config->iconpath);
				char* path = sstrcat(svn_path, "/README");
				FILE* readme = fopen(path, "r");
				char* content = NULL;
				if(readme != NULL){
					stat(path, &s);
					content = malloc(s.st_size + 1);
					content[s.st_size] = 0;
					fread(content, s.st_size, 1, readme);
					printf("		<description>%s: ~%s/%s: %s</description>\n", config->title, username, reponame, content);
					fclose(readme);
				}
				free(path);
				printf("	</channel>\n");
				if(content != NULL){
					printf("<item>\n");
					printf("	<title>README</title>\n");
					printf("	<link>%s%s/~%s/%s/README</link>\n", config->url_prefix, getenv("SCRIPT_NAME"), username, reponame);
					printf("	<dc:creator>%s</dc:creator>\n", username);
					printf("	<description>%s</description>\n", content);
					printf("</item>\n");
					free(content);
				}
				stat(revision_file, &s);
				char* crev = malloc(s.st_size + 1);
				crev[s.st_size] = 0;
				fread(crev, s.st_size, 1, f);
				int rev = atoi(crev);
				for(i = rev; i > 0; i--){
					char* num = malloc(1025);
					memset(num, 0, 1025);
					sprintf(num, "%d", i);
					printf("<item>\n");
					int* pipes = malloc(sizeof(*pipes) * 2);
					pipe(pipes);
					pid = fork();
					if(pid == 0){
						close(pipes[0]);
						int nullfd = open("/dev/null", O_RDWR);
						dup2(pipes[1], fileno(stdout));
						dup2(nullfd, fileno(stderr));
						chdir(svn_path);
						execl(config->svn, config->svn, "log", "-r", num, NULL);
						_exit(-1);
					}else{
						close(pipes[1]);
						waitpid(pid, 0, 0);
						char ch;
						int newline = 0;
						char* buf = malloc(1);
						buf[0] = 0;
						while(0 != read(pipes[0], &ch, 1)){
							char cond = newline < 3;
							if(ch != '\n' && cond == 1){
								continue;
							}else if(ch == '\n' && cond == 1){
								newline++;
								continue;
							}
							if(ch == '<'){
								char* tmp = buf;
								buf = sstrcat(tmp, "&lt;");
								free(tmp);
							}else if(ch == '>'){
								char* tmp = buf;
								buf = sstrcat(tmp, "&gt;");
								free(tmp);
							}else if(ch == '&'){
								char* tmp = buf;
								buf = sstrcat(tmp, "&amp;");
								free(tmp);
							}else if(ch == '\''){
								char* tmp = buf;
								buf = sstrcat(tmp, "&apos;");
								free(tmp);
							}else if(ch == '"'){
								char* tmp = buf;
								buf = sstrcat(tmp, "&quot;");
								free(tmp);
							}else{
								char* tmp = buf;
								char* buffer = malloc(2);
								buffer[1] = 0;
								buffer[0] = ch;
								buf = sstrcat(tmp, buffer);
								free(tmp);
								free(buffer);
							}
						}
						fflush(stdout);
						int j;
						newline = 0;
						for(j = strlen(buf) - 2;; j--){
							if(buf[j] == '\n'){
								newline++;
							}
							buf[j] = 0;
							if(newline == 1) break;
						}
						printf("	<link>%s%s/~%s/%s</link>\n", config->url_prefix, getenv("SCRIPT_NAME"), username, reponame);
						printf("	<title>Revision %d - %s</title>\n", i, buf);
						fflush(stdout);
						free(buf);
					}
					free(pipes);
					printf("	<description>\n");
					fflush(stdout);
					pipes = malloc(sizeof(*pipes) * 2);
					pipe(pipes);
					pid = fork();
					if(pid == 0){
						close(pipes[0]);
						int nullfd = open("/dev/null", O_RDWR);
						dup2(pipes[1], fileno(stdout));
						dup2(nullfd, fileno(stderr));
						chdir(svn_path);
						execl(config->svn, config->svn, "log", "--verbose", "-r", num, NULL);
						_exit(-1);
					}else{
						close(pipes[1]);
						char ch;
						int newline = 0;
						char* buf = malloc(1);
						buf[0] = 0;
						int counter = 0;
						char truncated = 0;
						while(1){
							int status = read(pipes[0], &ch, 1);
							if(status == 0 && kill(pid, 0) == 0) break;
							if(ch != '\n' && newline == 0){
								continue;
							}else if(ch == '\n' && newline == 0){
								newline = 1;
								continue;
							}
							counter++;
							if(ch == '<'){
								char* tmp = buf;
								buf = sstrcat(tmp, "&lt;");
								free(tmp);
							}else if(ch == '>'){
								char* tmp = buf;
								buf = sstrcat(tmp, "&gt;");
								free(tmp);
							}else if(ch == '&'){
								char* tmp = buf;
								buf = sstrcat(tmp, "&amp;");
								free(tmp);
							}else if(ch == '\''){
								char* tmp = buf;
								buf = sstrcat(tmp, "&apos;");
								free(tmp);
							}else if(ch == '"'){
								char* tmp = buf;
								buf = sstrcat(tmp, "&quot;");
								free(tmp);
							}else{
								char* tmp = buf;
								char* buffer = malloc(2);
								buffer[1] = 0;
								buffer[0] = ch;
								buf = sstrcat(tmp, buffer);
								free(tmp);
								free(buffer);
							}
							if(counter == 1024){
								char* tmp = buf;
								buf = sstrcat(tmp, " ... Truncated!");
								free(tmp);
								truncated = 1;
								break;
							}
						}
						close(pipes[0]);
						waitpid(pid, 0, 0);
						fflush(stdout);
						if(truncated == 0){
							int j;
							for(j = strlen(buf) - 2; buf[j] != '\n'; j--){
								buf[j] = 0;
							}
						}
						printf("%s\n", buf);
						fflush(stdout);
						free(buf);
					}
					printf("	</description>\n");
					printf("	<dc:date>\n");
					fflush(stdout);
					pid = fork();
					if(pid == 0){
						chdir(svn_path);
						execl(config->svn, config->svn, "info", "-r", num, "--show-item", "last-changed-date", NULL);
						_exit(-1);
					}else{
						waitpid(pid, 0, 0);
					}
					fflush(stdout);
					printf("	</dc:date>\n");
					printf("	<dc:publisher>\n");
					fflush(stdout);
					pid = fork();
					if(pid == 0){
						chdir(svn_path);
						execl(config->svn, config->svn, "info", "-r", num, "--show-item", "last-changed-author", NULL);
						_exit(-1);
					}else{
						waitpid(pid, 0, 0);
					}
					fflush(stdout);
					printf("	</dc:publisher>\n");
					printf("</item>\n");
					free(num);
					free(pipes);
				}
				free(crev);
				printf("</rdf:RDF>\n");
				fclose(f);
			}
			free(svn_path);
			free(revision_file);
		}
		free(_cutfrom);
	}else{
		printf("Status: 400 Bad Request\r\n");
		printf("Content-Type: text/plain\r\n");
		printf("\r\n");
		printf("Missing required parameters");
	}
}
